

module.exports = function (config, callback) {

    function toString(obj) {
        return Object.prototype.toString.call(obj);
    }

    function isFunction(obj) {
        return typeof obj === 'function';
    }

    function isObject(obj) {
        return typeof obj === 'object';
    }

    function isString(obj) {
        return typeof obj === 'string';
    }

    function isArray(obj) {
        return isObject(obj) && toString(obj) === '[object Array]';
    }

    function iterate(array, fn) {
        var length = array.length;
        for (var i = 0; i < length; ++i) {
            if (fn(array[i])) {
                return array[i];
            }
        }
    }

    function warn(error) {
        console.warn(error);
    }

    function throwError(error) {
        error = error instanceof Error ? error : new Error(error);
        if (callback) {
            callback(error);
        } else {
            warn(error);
        }
    }

    function setDefaults(shimLike) {
        var shim = shimLike;
        shim.countdown = 0;
        shim.subscribers = [];
        shim.finished = false;
        shim.processed = false;
        shim.initialized = false;
        return shim;
    }

    function normalizeString(source) {
        // @disable-flow
        return setDefaults(require('./shims/' + source));
    }

    function normalizeShimLike(source) {
        return setDefaults(source);
    }

    function normalize(source) {
        if (isString(source)) {
            return normalizeString(source);
        } else if (isObject(source) && isFunction(source.check) && isFunction(source.shim)) {
            return normalizeShimLike(source);
        } else {
            warn('Malformed shims config');
        }
    }

    var errors = [];
    var shimSet = [];
    var shimPairs = [];
    var remainedShims = 0;
    var progressShims = 0;
    var processing = true;

    function maybeFinish() {
        if (!processing && progressShims === 0) {
            var errorsMessage = '';
            iterate(errors, function (error) {
                errorsMessage += '\r\n' + error.toString();
            });
            if (remainedShims !== 0) {
                throwError('Shims have unsatisfied or circular dependencies' + errorsMessage);
            } else if (errors.length) {
                throwError(errorsMessage);
            } else if (callback) {
                callback();
            }
        }
    }

    function onProcess(err, shim) {
        if (err) {
            errors.push(err);
        } else {
            iterate(shim.subscribers, function (subscriber) {
                --subscriber.countdown;
                process(subscriber);
            });
        }
        shim.finished = true;
        progressShims--;
        remainedShims--;
        maybeFinish();
    }

    function process(shim) {
        if (shim.countdown || shim.processed) {
            return; // shim dependencies are not ready || shim was already processed
        }

        progressShims++; // atomic
        shim.processed = true;

        try {
            if (shim.check()) {
                shim.shim(function (err) {
                    onProcess(err, shim);
                });
            } else {
                onProcess(undefined, shim);
            }
        } catch (e) {
            errors.push(e);
            if (!shim.finished) {
                progressShims--;
            }
        }
    }

    iterate(config, function (source) {
        var shim = void 0,
            dependencies = [];

        if (isArray(source)) {
            shim = normalize(source[0]);

            if (isArray(source[1])) {
                // @disable-flow
                iterate(source[1], function (source) {
                    var dependency = normalize(source);
                    if (dependency) {
                        dependencies.push(dependency);
                    }
                });
            }
        } else {
            shim = normalize(source);
        }

        if (shim) {
            shimPairs.push([shim, dependencies]);
        }
    });

    var duplicate = iterate(shimPairs, function (pair) {
        var shim = pair[0];
        var dependencies = pair[1];
        if (shim.initialized) {
            return true; // set should be unique
        }
        shim.initialized = true;
        iterate(dependencies, function (dependency) {
            shim.countdown++;
            dependency.subscribers.push(shim);
        });
        shimSet.push(shim);
    });

    if (duplicate) {
        return throwError('Shims config contains duplicates');
    }

    remainedShims = shimSet.length;
    processing = true;
    iterate(shimSet, function (shim) {
        process(shim);
    });
    processing = false;

    maybeFinish();
};